

from pandas import Series, DataFrame
import pandas as pd
from collections import OrderedDict
import warnings
from .read.readmdb import ReadMdb


class MapTables:
    """
    Tables with data related to vegetation map elements

    Methods
    -------
        get_vegtype
            Return vegetation type for each mapped element.
        get_mapspecies
            Return species data attached to mapped elements.
        get_abiotiek
            Return environmental field observatons attached to mapped 
            elements.
        get_pointspecies
            Return point locations for mapped plant species.
        get_year
            Return year of mapping, returns 0000 if no dates are present.

    Classmethods
    ------------
        from_mdb
            Create MapTables object from Microsoft Access mdb filepath.

    Notes
    -----
    A mapped element can be a polygon or a line. Spatial data for these
    elements are stored in shapefiles and linked to the table data by 
    the attribute ElmID.
    
    """

    _DS_tablenames = ['vegetatietype','sbbtype']

    _mapping_colnames = OrderedDict({
        'Element' : {
            'intern_id' : 'locatie_id',
            },
        'KarteringVegetatietype' : {
            'locatie': 'locatie_id',
            'vegetatietype': 'vegtype_code',
            'bedekking':'vegtype_bedekkingcode',
            'bedekking_num':'vegtype_bedekkingnum',
            },
        'VegetatieType' : {
            'typenummer':'vegtype_nr',
            'code':'vegtype_code',
            'gemeenschap':'vegtype_naam',
            'vorm':'vegtype_vorm',
            'sbbtype':'sbbcat_id',
            'sbbtype2':'sbbcat2_id',
            'opmerking':'vegtype_note',
            },
        'SbbType': {
            'cata_id':'sbbcat_id',
            'versie':'sbbcat_versie',
            'code':'sbbcat_code',
            'klassenaamned':'sbbcat_klassenaam',
            'verbrgnaamned':'sbbcat_kortenaam',
            'asscocrgnaamned':'sbbcat_assrgnaam',
            'subassocnaamned':'sbbcat_subassnaam',
            'landtypened':'sbbcat_nednaam',
            'landtypewet':'sbbcat_wetnaam',
            'vervallen':'sbbcat_vervallen',
            'vervangbaarheid':'sbbcat_vervangbaarheid',
            },
        'KarteringSoort': {
            'locatie':'locatie_id', 
            'soortcode':'krtsrt_srtcode', 
            'bedekking':'krtsrt_bedcode', 
            'aantalsklasse':'krtsrt_aantalsklasse', 
            'bedekking_num':'krtsrt_bednum',
            },
        'CbsSoort':{
            'soortnr':'cbs_srtcode',
            'floron':'cbs_floron',
            'wetenschap':'cbs_srtwet',
            'nederlands':'cbs_srtned',
            'zeldzaamheidsklasse':'cbs_zeldzaamheid',
            'trendklasse':'cbs_trend',
            'rl2000':'cbs_rl2000',
            'rl2000kort':'cbs_rl2000kort',
            },
        'PuntLocatieSoort': {
            'id':'pntid',
            'loctype':'pntloctype',
            'x_coord':'xcr',
            'y_coord':'ycr',
            'groep':'srtgroep',
            'nummer':'srtnr',
            'naam':'srtnednaam',
            'wetens':'srtwetnaam',
            'sbb_kl':'srtsbbkl',
            'tansley':'srttansley',
            'datum':'srtdatum',
            'waarn':'srtwrnmr',
            'opm':'srtopm',
            },
        'KarteringAbiotiek': {
            'locatie':'locatie_id',
            'abiotiek':'abio_code',
            },
        'Abiotiek': {
            'code':'abio_code',
            'omschrijving':'abio_wrn',
            },
        })


    def __init__(self,tables=None,mdbpath=None):
        """
        Parameters
        ----------
        tables : OrderedDict
            Dictionary of tables from mdb file
        mdbpath : string
            Filepath to mdb sourcefile (for userwarnings).

        Notes
        -----
        Use classmethod MapTables.from_mdb(<filepath>) to create a 
        MapTables instance with data from a Microsoft Access mdb file.
        """

        self._tbldict = tables
        if mdbpath is None:
            mdbpath = ''
        self._filepath = mdbpath

        # numeric to string type
        self._tbldict['Element'] = self._tbldict['Element'].astype(
            {'locatie_id':str,'elmid':str,})
        self._tbldict['KarteringVegetatietype']=self._tbldict['KarteringVegetatietype'].astype(
            {'locatie_id':str,})
        self._tbldict['VegetatieType'] = self._tbldict['VegetatieType'].astype(
            {'sbbcat_id': str, 'sbbcat2_id': str,}) 
        self._tbldict['SbbType'] = self._tbldict['SbbType'].astype(
            {"sbbcat_id": str, "sbbcat_versie": str, "sbbcat_vervangbaarheid": str,}) 
        self._tbldict['KarteringSoort']=self._tbldict['KarteringSoort'].astype(
            {'locatie_id':str,'krtsrt_srtcode':str,})
        self._tbldict['CbsSoort']=self._tbldict['CbsSoort'].astype(
            {'cbs_srtcode':str,})
        self._tbldict['KarteringAbiotiek']=self._tbldict['KarteringAbiotiek'].astype(
            {'locatie_id':str,})


        # change vevangbaarheid 5.0 to 5 stingtype
        self._tbldict['SbbType']['sbbcat_vervangbaarheid']=self._tbldict['SbbType']['sbbcat_vervangbaarheid'].str[:1]

        # convert column locatietype to lowercase
        # (locatietype can be: 'v','l','V','L')
        self._tbldict['Element']['locatietype'] = self._tbldict['Element']['locatietype'].str.lower()

        # fix small errors that occur in just a few (or just one) mdbfiles
        # smallfix01
        colnames = self._tbldict['Element'].columns
        if ((not 'sbbtype' in colnames) and ('sbbtype1' in colnames)):
            self._tbldict['Element']=self._tbldict['Element'].rename(
                columns={'sbbtype1':'sbbtype'})


    def __repr__(self):
        return f'MapTables (n={self.__len__()})'

    def __len__(self):
        return len(self._tbldict['Element'])

    @classmethod
    def from_mdb(cls,filepath):
        """
        Create MapTables object from Microsoft Access mdb filepath."

        Parameters
        ----------
        filepath : str
            valid filepath to Microsoft Access mdb file
        """

        if not isinstance(filepath,str):
            fptype = type(filepath)
            raise ValueError (f'Parameter filepath must be type "str" '
                f'not type {fptype}.')

        # open mdb file and check format is Digitale Standaard
        mdb = ReadMdb(filepath)

        if not mdb.all_tables():
            msg = f'{filepath} is not a valid Digitale Standaard database.'
            ##raise Exception(msg)
            warnings.warn(msg)
            return None
            
        #if any([item in mdb.tablenames() for item in cls._DS_tablenames]):
        #    raise Exception(f'{mdb} is not a valid Digitale Standaard database.')

        # all mdb tables to dict
        mdbtables = mdb.all_tables()
        maptables = {}
        for tblname in mdbtables.keys():
            mdbtbl = mdbtables[tblname]
            mdbtbl.columns = map(str.lower,mdbtbl.columns)
            if tblname in cls._mapping_colnames.keys():
                mdbtbl = mdbtbl.rename(columns=cls._mapping_colnames[tblname])
            maptables[tblname] = mdbtbl

        return cls(tables=maptables,mdbpath=filepath)


    def get_vegtype(self,loctype='v',select='all'):
        """
        Return vegetation type for each mapped element.

        Parameters
        ----------
        loctype : {'v','l'}, default 'l'
            Element location type
        select : {'all','maxcov'}, default 'all'
            Select from multiple instances of polygon.
            maxcov : select vegetation type with largest numeric cover.
        Notes
        -----
        A mapped element can have multiple vegetation types. Therefore 
        the table returned can have multiple rows with the same value 
        for the map polygon id elmid. 

        The fields bedekking and bedekking_num show the cover a 
        vegetation type has within a map polygon.
        """
        if loctype not in ['v','l']:
            warnings.warn((f'Loctype must be "v" of "l", not {loctype}. '
                f'Elements of loctyp "v" will be returned.'))
            loctype = 'v'


        elmcolnames = ['locatie_id', 'elmid', 'locatietype', 'datum']
        element = self._tbldict['Element'][elmcolnames]
        isloctype = element['locatietype']==loctype
        element = element[isloctype].copy()

        vegloc = self._tbldict['KarteringVegetatietype']
        element = pd.merge(element,vegloc,left_on='locatie_id',
            right_on='locatie_id',suffixes=(None,'_vegloc'),
            validate='one_to_many')

        vegtype = self._tbldict['VegetatieType']
        element = pd.merge(element,vegtype,left_on='vegtype_code',
            right_on='vegtype_code',suffixes=(None,'_vegtype'),
            validate='many_to_one')

        sbbtype = self._tbldict['SbbType']
        element = pd.merge(element,sbbtype,left_on='sbbcat_id',
            right_on='sbbcat_id',suffixes=(None,'sbbtype'),
            validate='many_to_one')

        element['datum'] = element['datum'].apply(lambda x: x.strftime(
            '%d%m%Y') if not pd.isna(x) else '')

        colnames = ['elmid','datum','locatietype','vegtype_code',
            'vegtype_naam','vegtype_vorm','vegtype_bedekkingcode',
            'vegtype_bedekkingnum',
            'sbbcat_code', 'sbbcat_wetnaam','sbbcat_nednaam',
            'sbbcat_kortenaam','sbbcat_vervangbaarheid']
        element = element[colnames].copy()

        if select=='maxcov':
            element = element.sort_values(['elmid',
                'vegtype_bedekkingnum'],ascending=False)
            element = element.groupby('elmid').head(1)

        return element.copy()

    def get_pointspecies(self):
        """Return table of point locations for mapped plant species"""
        pntsrt = self._tbldict['PuntLocatieSoort'].copy()
        pntsrt['srtdatum'] = pntsrt['srtdatum'].apply(
            lambda x: x.strftime('%d%m%Y') if not pd.isna(x) else '')
        return pntsrt

    """
    def get_year(self):
        Return year of mapping, returns 0000 if no dates are present
        dates = pd.to_datetime(self._tbldict['Element']['datum'],errors='coerce')
        years = list(dates.dt.year.unique())

        if len(years)==1:
            if pd.isnull(years[0]):
                return '0000'
            return str(years[0])
        return str(min(years))+'-'+str(max(years))
    """

    def get_years(self):
        """Return frequency of mapped elements by year."""
        dates = pd.to_datetime(self._tbldict['Element']['datum'],errors='coerce')
        years = dates.dt.year.value_counts(sort=False)

        if years.empty:
            warnings.warn((f'No valid dates in {self._filepath}.'),stacklevel=1)
            ##return None

        return years

    def get_mapspecies(self,loctype='all'):
        """Return species data attached to mapped elements.

        Parameters
        ----------
        loctype : {'all','v','l'}, default 'all'
            Results contain only values for this location type.

        Returns
        -------
        pandas DataFrame

        Note
        ----
        Multiple species can be mapped for a map element. Values for
        field ElmID will therefore not be unique.
        """

        if loctype not in ['all','v','l']:
            raise ValueError(f'Invalid loctype {loctype}')

        elmcolnames = ['locatie_id', 'elmid', 'locatietype', 'datum','sbbtype']
        element = self._tbldict['Element'][elmcolnames]

        krtsrt = self._tbldict['KarteringSoort']
        mapspec = pd.merge(element,krtsrt,left_on='locatie_id',
            right_on='locatie_id',how='right',suffixes=(None,'_krtsrt'),
            validate='one_to_many')

        cbscolnames = ['cbs_srtcode','cbs_srtwet','cbs_srtned',]
        cbs = self._tbldict['CbsSoort'][cbscolnames]
        mapspec = pd.merge(mapspec,cbs,left_on='krtsrt_srtcode',
            right_on='cbs_srtcode',how='left',suffixes=(None,'_cbs'),
            validate='many_to_one')

        mapspec = mapspec.drop(columns=['locatie_id','cbs_srtcode'])

        if loctype in ['v','l']:
            mapspec = mapspec[mapspec['locatietype']==loctype]

        return mapspec


    def get_abiotiek(self,loctype='all'):
        """Return table of environmental observatons"""

        if loctype not in ['all','v','l']:
            raise ValueError(f'Invalid loctype {loctype}')

        elmcolnames = ['locatie_id', 'elmid', 'locatietype', 'datum']
        element = self._tbldict['Element'][elmcolnames]
        abi = self._tbldict['KarteringAbiotiek']
        mapabi = pd.merge(element,abi,left_on='locatie_id',
            right_on='locatie_id',how='left',suffixes=(None,'_abi'),
            validate='one_to_many')

        abicode = self._tbldict['Abiotiek']
        mapabi = pd.merge(mapabi,abicode,left_on='abio_code',
            right_on='abio_code',how='left',suffixes=(None,'_abicode'),
            validate='many_to_one')

        if loctype in ['v','l']:
            mapabi = mapabi[mapabi['locatietype']==loctype]

        mapabi = mapabi.drop(columns=['locatie_id'])
        return mapabi[mapabi['abio_code'].notnull()]















import warnings
from pandas import Series, DataFrame
import pandas as pd
import geopandas as gpd

from .read.readshapefile import ReadShapeFile

class MapElements:
    """Spatial data for mapped elements  
    
    Methods
    -------
    get_shape : GeoDataFrame
        Return table of spatial data.
    get_colnames : list
        Return attribute column names.
    get_geomtype : str
        Return geometry type (vlakken or lines)
    get_filepath : str
        Return sourcefile path
    get_boundary : shape
        Return outer boundary of mappend area

    Classmethods
    ------------
    from_shapefile : MapElements
        Create MapElements instance from shapefilepath.
    """
    
    def __init__(self,shape=None,filepath=None):
        """MapElements constructor
        
        Parameters
        ----------
        shape : GeoDataFrame
            Spatial data
        filepath : str
            Sourcefile path (used for warnings)
            
        """
        if shape is None:
            shape = gpd.GeoDataFrame()
        self._shape = shape
        self._filepath = filepath

        if not self._shape.empty:
            self._geom_type = list(set(self._shape.geom_type))[0].lower()

            if self._shape.crs is None:
                self._shape = self._shape.set_crs('epsg:28992')


    def __repr__(self):
        return self._shape.__repr__()
        
    def __len__(self):
        return self._shape.__len__()

    def get_shape(self):
        """Return spatial data"""
        return self._shape
        
    def get_colnames(self):
        return list(self._shape)
        
    def get_geomtype(self):
        return self._geom_type
        
    def get_filepath(self):
        return self._filepath

    def get_boundary(self):
        """Return outer boundary of mapped area"""
        outline = self._shape.copy()
        outline['diss']=1
        outline = outline.dissolve('diss')
        outline = outline.geometry.boundary
        return outline
        

    @classmethod
    def from_shapefile(cls,filepath):
        """
        Create MapElements object from ESRI shapefile filepath."

        Parameters
        ----------
        filepath : str
            valid filepath to ESRI shapefile
        """

        if not isinstance(filepath,str):
            fptype = type(filepath)
            raise ValueError (f'Parameter filepath must be type "str" '
                f'not type {fptype}.')

        # open shapefile and check presence of column ElmID
        shp = ReadShapeFile(filepath)
        
        return cls(shape=shp._shape,filepath=filepath)




from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="DSreader",
    version="0.0.1",
    description="Read data from Digital Standard vegetation mapping projects",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tdmeij/DSreader",
    author="Thomas de Meij",
    author_email="thomasdemeij@gmail.com",
    license="MIT",
    packages=["DSreader"],
    install_requires=[
        'pandas','numpy','geopandas','fiona','json','pyodbc','plotly',
        ],
    include_package_data=True,
    package_data={'': ['data/*.csv']},
)



"""
DSreader is a python package for reading data from vegetation maps 
organised according to the Digital Standard.

The digital Standard is a dataformat developed in the nineties
by the Dutch National Forestry Service (Staatsbosbeheer). A vegetion
map comprises an ESRI shapefile with polygons and an Microsoft Access 
database with information about the content of the polygons.

Main package classes
--------------------
MapData
    Manage vegetion map data and connect spatial data to non-spatial
    data.

Helper classes
--------------
ListProjects
    Create table of filenames organised by project under a given root file
MapElements
    Manage spatial data for map elements.
MapTables
    Manage tables with non-spatial data related to mapped elements.
ReadMdb
    Read Microsoft Access .mdb file and extract tables as pd.DataFrame.
ReadShapeFile
    Read ESRI Shapefile and correct errors if necessary.

Classes for spatial analysis
----------------------------
SampleShape
    Create regular grid and sample polygon map at grid points.
SankeyTwoMaps
    Plot Sankey diagram comparing changes in management quality between 
    two maps.
   
"""

# import classes
from .maptables import MapTables
from .mapelements import MapElements
from .mapdata import MapData
from .read.readshapefile import ReadShapeFile
from .read.readmdb import ReadMdb
from .tools.sampleshape import SampleShape
from .tools.listprojects import ListProjects
from .plot.sankey_two_maps import SankeyTwoMaps
from .data import data

# import functions
from .tools.filetools import relativepaths as relativepaths
from .tools.filetools import absolutepaths as absolutepaths




import os
from collections import OrderedDict
import warnings
import numpy as np
from pandas import Series, DataFrame
import pandas as pd
from shapely.geometry import Point
import geopandas as gpd

from .maptables import MapTables
from .mapelements import MapElements

class MapData:
    """
    Vegetion map data connectting spatial and non-spatial data.
    
    Methods
    -------
    get_polys()
        Return polygon geometry as GeoPandas dataframe.
    get_lines()
        Return line geometry as GeoPandas dataframe.
    get_vegtype()
        Return mapped polygons with vegetation type.
    get_mapspecies(self,loctype='v'):
        Return map polygons with species data.
    get_pointspecies()
        Return species point data.
    get_abiotiek()
        Return environmental observations.
    get_boundary()
        Return outer boundary of mapped area.
    to_shapefile()
        Save table to ESRI shapefile.

    Classmethods
    ------------
    from_filepaths()
        Create MapData instance from filepaths.
        
    """

    _shapefile_colnames = {
        'vegtype': OrderedDict(
             elmid='elmid',
             datum='datum',
             locatietype='loctype',
             vegtype_code='veg_code',
             vegtype_naam='veg_naam',
             vegtype_vorm='veg_vorm',
             vegtype_bedekkingcode='veg_bedcod',
             vegtype_bedekkingnum='veg_bednum',
             sbbcat_code='sbb_code',
             sbbcat_wetnaam='sbb_wetnm',
             sbbcat_nednaam='sbb_nednm',
             sbbcat_kortenaam='sbb_kortnm',
             sbbcat_vervangbaarheid='sbb_vvangb',
             oppha='oppha',
             geometry='geometry',
        ),
        'mapspecies': OrderedDict(
            elmid='elmid',
            locatietype='loctype', 
            datum='datum', 
            sbbtype='sbbtype',
            krtsrt_srtcode='srtcode', 
            krtsrt_bedcode='bedcode', 
            krtsrt_aantalsklasse='aantkl',
            krtsrt_bednum='bednum', 
            cbs_srtwet='srtwet', 
            cbs_srtned='srtned',
            oppha='oppha', 
            geometry='geometry', 
        ),
        'pointspecies' : OrderedDict(
            pntid='pntid', 
            pntloctype='loctype', 
            srtgroep='srtgroep', 
            srtnr='srtnr', 
            srtnednaam='nednaam',
            srtwetnaam='wetnaam', 
            srtsbbkl='sbbkl', 
            srttansley='tansley', 
            srtdatum='datum', 
            srtwrnmr='wrnmr',
            srtopm='opm', 
            xcr='xcr', 
            ycr='ycr', 
            geometry='geometry',
        ),
        'abiotiek' : OrderedDict(
            elmid='elmid',
            oppha='oppha', 
            locatietype='loctype', 
            datum='datum',
            abio_code='abio_code', 
            abio_wrn='abio_wrn',
            geometry='geometry',
        ),
    } # shapefile maximum column width is 10


    def __init__(self,maptables=None,polygons=None,lines=None):
        """MapPolygons constructor.

        Parameters
        ----------
        maptbl : DSreader.MapTables
            Map table data
        polygons : DSreader.MapElements
            Spatial data for all map polygons
        lines : DSreader.MapElelemnts
            Spatial data for all map lines
        """
        if maptables is None:
            maptables = MapTables()
        self._maptbl = maptables

        if polygons is None:
            polygons = MapElements()
        self._mapelements_polygons = polygons

        if lines is None:
            lines = MapElements()
        self._mapelements_lines = lines            
            
        self._maptblpath = self._maptbl._filepath
        self._polypath = self._mapelements_polygons._filepath
        self._linepath = self._mapelements_lines._filepath

        self._poly = self._mapelements_polygons.get_shape()
        if not self._poly.empty:
            self._poly = self._poly[['elmid','geometry']].copy()
            self._poly = self._poly.astype({'elmid':'str'})
            self._poly['oppha']=self._poly['geometry'].area/10000

        self._lines = self._mapelements_lines.get_shape()
        if not self._lines.empty:
            self._lines = self._lines[['elmid','geometry']].copy()
            self._lines = self._lines.astype({'elmid':'str'})

    def __repr__(self):
        return f'MapData (n={self._maptbl.__len__()})'

    @classmethod
    def from_filepaths(cls,mdbpath=None,polypath=None,linepath=None):
        """Create MapData instance from filepaths

        
        Parameters
        ----------
        mdbpath : str, optional
            Valid filepath for Microsoft Access mdb-file.
        polypath : str, optional
            Valid filepath to ESRI shapefile with vegetation polygons.
        linepath : str, optional
            Valid filepath to ESRI shapefile with vegetation lines.


        Returns
        -------
        MapData instance.
        
        
        """
        if isinstance(mdbpath,str):
            tables = MapTables.from_mdb(mdbpath)
        else:
            tables = MapTables()

        if isinstance(polypath,str):
            poly = MapElements.from_shapefile(polypath)
        else:
            poly = MapElements()

        if isinstance(linepath,str):
            line = MapElements.from_shapefile(linepath)
        else:
            line = MapElements()

        return cls(maptables=tables,polygons=poly,lines=line)


    def get_polys(self):
        """Return polygon geometry as GeoPandas dataframe"""
        return self._poly

    def get_lines(self):
        """Return line geometry as GeoPandas dataframe"""
        return self._lines


    def get_vegtype(self,loctype='v',select='all'):
        """
        Return mapped polygons with vegetation type
        
        Parameters
        ----------
        element : {'v','l'}, default ' v'
            Map element type.
        select : {'all','maxcov'}, default 'all'
            Select from multiple instances of polygon.
            maxcov : select vegetation type with largest numeric cover.

        Returns
        -------
        pd.DataFrame


        """
        if loctype=='v':
            shape = self._poly
            shapepath = self._polypath
        if loctype=='l':
            shape = self._lines
            shapepath = self._linepath

        vegtbl = self._maptbl.get_vegtype(loctype=loctype,select=select)

        if vegtbl.empty:
            name = str(vegtbl)
            warnings.warn((f'Empty vegetation data in {name}'))
            return DataFrame()

        try:
            shape = pd.merge(shape,vegtbl,how='left',left_on='elmid',right_on='elmid',
                ) #validate='one_to_many') TURNED OFF BECAUSE OF DUPLICATE ELMIDS
                
        except Exception as e:
            warnings.warn((f'Merge caused fatal exception: "{e}" '
                f'on shapefile {shapepath}" '
                f'and Access database "{self._maptblpath}"'))
            shape = DataFrame()

        else:
            shape = shape.dropna(subset=['locatietype'])

        return shape

    def get_mapspecies(self,loctype='v'):
        """Return map polygons with species data


        Parameters
        ----------
        element : {'v','l'}, default ' v'
            Map element type.

        Returns
        -------
        pd.DataFrame


        """
        if loctype=='v':
            shape = self._poly
        if loctype=='l':
            shape = self._lines
            
        mapspec = self._maptbl.get_mapspecies(loctype=loctype)

        mapspec = pd.merge(shape,mapspec,how='outer',left_on='elmid',
            right_on='elmid',validate='many_to_many')

        mapspec = mapspec.dropna(subset=['locatietype'])

        return mapspec


    def get_pointspecies(self):
        """Return species point data"""
        spc = self._maptbl.get_pointspecies()
        if not spc.empty:
            geometry = [Point(x,y) for x,y in zip(spc.xcr, spc.ycr)]
            spc = gpd.GeoDataFrame(spc,crs="EPSG:28992",geometry=geometry)

        return spc


    def get_abiotiek(self,loctype='v'):
        """Return environmental observations"""
        if loctype not in ['v','l']:
            warnings.warn((f'Invalid loctype {loctype}, '
                f'abiotiek for loctype "v" will be returned.'))
            loctype='v'

        if loctype=='v':
            shape = self._poly
            shppath = self._polypath
        if loctype=='l':
            shape = self._lines
            shppath = self._linepath

        abi = self._maptbl.get_abiotiek(loctype=loctype)

        #try:
        abi = pd.merge(shape,abi,how='left',left_on='elmid',
            right_on='elmid',validate='many_to_many') 
                
        #except Exception as e:
        #    warnings.warn((f'Merge caused fatal exception: "{e}" '
        #        f'on shapefile {shppath}" '
        #        f'and Access database "{self._maptblpath}"'))
        #    abi = DataFrame()

        #else:
        abi = abi.dropna(subset=['locatietype'])

        return abi

    def get_boundary(self):
        """Return outer boundary of mapped area"""
        return self._mapelements_polygons.get_boundary()

    def to_shapefile(self,tablename=None,loctype='v',filepath=None):
        """Save table to ESRI shapefile

        Parameters
        ----------
        tablenamne : {'vegtype','mapspecies','pointspecies'}
            Kinde of table to save
        loctype : {'v','l'}
            map element type
        filepath : str
            Valid filepath for shapefile

        Returns
        -------
        pd.DataFrame

        Notes
        -----
        ESRI shapefile column names have a maximum length of ten
        characters and datetimes can not be saved, only dates.
        The returned value is the table that has been saved to 
        shapefile or it is an empty DataFrame.
        """
        # validate tablename
        tablenames = ['vegtype','mapspecies','pointspecies','abiotiek']
        if tablename not in tablenames:
            warnings.warn((f'{tablename} is not a valid tablename. '
                f'No file has been saved.'))
            return DataFrame()

        # validate element
        if loctype not in ['v','l']:
             warnings.warn((f'{loctype} is not a valid element type. '
                f'Elements of type "v" will be saved.'))           
             loctype='v'

        # validate filepath and correct
        dirname = os.path.dirname(filepath)
        if (dirname!='') and (not os.path.exists(dirname)):
            warnings.warn((f'{dirname} is not a valid directory. '
                f'No file has been saved.'))
            return DataFrame()
        if os.path.splitext(filepath)[1]=='':
            filepath = filepath+'.shp'
        if os.path.splitext(filepath)[1]!='.shp':
            filepath = os.path.splitext(filepath)[0]+'.shp'

        # get the right table 
        if tablename=='vegtype':
            table = self.get_vegtype(loctype=loctype)
        elif tablename=='mapspecies':
            table = self.get_mapspecies(loctype=loctype)
        elif tablename=='pointspecies':
            table = self.get_pointspecies()
        elif tablename=='abiotiek':
            table = self.get_abiotiek(loctype=loctype)
        else:
            raise ValueError('{tablename} is not a valid table name.')

        if not table.empty:

            # rename columns
            table = table.rename(columns=self._shapefile_colnames[tablename])

            # date to string
            if 'datum' in table.columns: # shapefile has no datetime type
                all_strings = all((v is np.nan) or isinstance(v, str) 
                    for v in table['datum'])
                if all_strings:
                    table['datum'] = table['datum'].fillna('')
                else:
                    table['datum'] = table['datum'].apply(
                        lambda x: x.strftime(
                        '%d%m%Y') if not pd.isna(x) else '')

            # check if all columns are present
            shapecols = self._shapefile_colnames[tablename].values()
            coldif = set(table.columns)-set(shapecols)
            if len(coldif)!=0:
                warnings.warn((f'Unknown column names in table '
                    f'{tablename}: {coldif} in {filepath}.'))

            coldif2 = set(shapecols) - set(table.columns)
            if len(coldif2)!=0:
                warnings.warn((f'Missing column names in table '
                    f'{tablename}: {coldif2}.'))

            # order columns
            ordered_colnames = [colname for colname in shapecols
                if colname in table.columns] + list(coldif)
            table = table[ordered_colnames].copy()

            # save table
            table.to_file(filepath)

        return table
